from .errors import InternalSeverError
from .serializer import JsonSerializer
from .response import ErrorResponse
import sys


class ExceptionHandler(object):

    def __init__(self, logger):
        self.logger = logger
        self.serializer = JsonSerializer()

    def handle_general_exception(self, url, method, e, request):
        """
        Logs an exception then returns an InternalServerError http response body
        """
        self.logger.error('Exception on {0} [{1}]'.format(url, method), exc_info=sys.exc_info())
        return self.handle_http_exception(InternalSeverError(), request)

    def handle_http_exception(self, http_error, request):
        """
        Builds a response body for a http exception.
        """
        error_response = ErrorResponse(http_error, request)
        serialized_response = self.serializer.serialize_response(error_response)
        return serialized_response








from . import errors, NotSet, fields, params


class ParametersMetaClass(type):

    def __new__(mcs, parameters_name, bases, attrs):
        params_dict = {}
        for name, param in list(attrs.items()):
            if isinstance(param, params.ResourceParam):
                param.name = name
                params_dict[name] = param
            elif isinstance(param, fields.ResourceField):
                raise Exception('Expected subclass of ResourceParam, got {0}'.format(param))
        attrs['_params'] = params_dict
        return super().__new__(mcs, parameters_name, bases, attrs)


class Parameters(object, metaclass=ParametersMetaClass):

    def __new__(cls, *args, **kwargs):
        return super().__new__(cls)

    @classmethod
    def validate(cls, input_params):
        params_dict = {}

        for name, param in input_params.items():
            if name not in cls._params:
                raise errors.ValidationError(name + ' is not a supported query parameter.')

        for name, param in cls._params.items():
            if name in input_params:
                params_dict[name] = param.validate(input_params[name])
            else:
                default = param.default()
                if default is not NotSet:
                    params_dict[name] = default
        return params_dict



from . import errors, validators, NotSet


class ResourceParam(object):
    validator_type = None
    order_counter = 0

    def __init__(self, default=NotSet, notnull=False, options=None, *args, **kwargs):
        self.flags = {'notnull': notnull, 'default': default, 'options': options, 'required': False}
        self.name = 'noname'

        self.order_value = ResourceParam.order_counter
        ResourceParam.order_counter += 1

        # a hook to allow subclasses to add their own unique parameters
        self.set_unique_attributes(**kwargs)

        # create validator
        self._validator = self._get_validator()
        self.flags['default'] = self._validator.validate(default, cast=False)

    def __str__(self):
        return '{0}:{1}'.format(self.__class__.__name__, self.name)

    def validate(self, value, cast=True):
        return self._validator.validate(value, cast=cast)

    def default(self):
        return self.flags['default']

    def set_unique_attributes(self, **kwargs):
        pass

    def _get_validator(self):
        if self.validator_type:
            vt_cls = self.validator_type
            return vt_cls(self)
        else:
            raise NotImplementedError('Base class ResourceParam has no validator')


class CharParam(ResourceParam):
    validator_type = validators.CharValidator

    def set_unique_attributes(self, max_length=None):
        self.flags['max_length'] = max_length

    def validate(self, value, cast=False):
        """
        Because strings are expected, we don't want to cast
        """
        return super().validate(value, cast)


class IntParam(ResourceParam):
    validator_type = validators.IntValidator


class DecimalParam(ResourceParam):
    validator_type = validators.DecimalValidator


class DateParam(ResourceParam):
    validator_type = validators.DateValidator


class DateTimeParam(ResourceParam):
    validator_type = validators.DateTimeValidator


class BoolParam(ResourceParam):
    validator_type = validators.BoolValidator


class ListParam(ResourceParam):
    validator_type = validators.ListValidator

    def set_unique_attributes(self, item_type=None):
        if item_type:
            if item_type == validators.NotSet or not isinstance(item_type, ResourceParam):
                raise errors.ValidationError('ListParam must have an item_type set to a valid ResourceParam')
            self.flags['item_type'] = item_type
        else:
            self.flags['item_type'] = None

class JSONParam(ResourceParam):
    """A JSON resource param type.

    :param schema: Optional dict with JSON schema as per json-schema.org
    """
    validator_type = validators.JSONValidator

    def set_unique_attributes(self, schema=None):
        if schema:
            self.flags['json_schema'] = schema



# -*- coding: utf-8 -*-

from . import errors, fields, NotSet


VALID_METHODS = {'get', 'post', 'put', 'patch', 'delete', 'options'}
VALID_QUERY_PARAMETERS = {'sort', 'offset', 'limit'}


class ResourceMetaClass(type):

    # Note: attrs comes in alphabetical order,
    # unsure how to maintain declared order of fields
    def __new__(mcs, resource_name, bases, attrs):
        attrs['_fields'] = {}
        # inherit fields from parent classes
        for base in bases:
            if hasattr(base, '_fields'):
                attrs['_fields'].update(base._fields)
        # update fields based on current class attributes
        attrs['_fields'].update(mcs._get_fields(bases, attrs))
        return super().__new__(mcs, resource_name, bases, attrs)

    # Note: will likely need some sort of sorted dictionary to maintain
    # field order
    @staticmethod
    def _get_fields(bases, attrs):

        # gather fields to add to _fields dictionary
        # also replaces field definition with a property to wrap the
        # _fields dictionary
        field_dict = {}
        for name, field in attrs.items():
            if isinstance(field, fields.ResourceField):
                field.name = name
                field_dict[name] = field
                attrs[name] = property(
                    fget=ResourceMetaClass._gen_get_prop(name),
                    fset=ResourceMetaClass._get_set_prop(name),
                )
        return field_dict

    @staticmethod
    def _gen_get_prop(name):
        def get_field_property(self):
            return self._get(name)
        return get_field_property

    @staticmethod
    def _get_set_prop(name):
        def set_field_property(self, value):
            return self._set(name, value)
        return set_field_property


class Resource(object, metaclass=ResourceMetaClass):
    _fields = {}

    def __new__(cls, *args, **kwargs):
        obj = super().__new__(cls)
        obj._values = {}
        for name, field in obj.all_fields():
            if callable(field.default):
                obj._values[name] = field.default()
            else:
                obj._values[name] = field.default
        return obj

    def __init__(self, *args, **kwargs):
        self._partial = getattr(self, '_partial', False)
        self._init(*args, **kwargs)

    @classmethod
    def partial(cls, *args, **kwargs):
        resource = cls.__new__(cls)
        resource._partial = True
        resource.clear()
        resource._init(*args, **kwargs)
        return resource

    # Is there a better way to do this?
    @classmethod
    def init_from_obj(cls, obj, partial=False, mapping=None, override=None,
                      cast=False):
        mapping = mapping if mapping else {}
        resource = cls.__new__(cls)
        resource._partial = partial
        if resource._partial:
            resource.clear()
        resource.from_obj(obj, mapping, override, cast)
        resource.validate()
        return resource

    @classmethod
    def init_from_dict(cls, data, partial=False, cast=False):
        resource = cls.__new__(cls)
        resource._partial = partial
        if resource._partial:
            resource.clear()
        resource.from_dict(data, cast)
        resource.validate()
        return resource

    @classmethod
    def empty(cls):
        resource = cls.__new__(cls)
        resource._partial = False
        resource.clear()
        return resource

    def _init(self, *args, **kwargs):
        if args and len(args) == 1 and isinstance(args[0], dict):
            self.from_dict(args[0])
        elif args:
            raise Exception(
                'Resource initiation accepts only a dictionary or fields by '
                'keyword.'
            )
        if kwargs:
            self.from_dict(kwargs)
        self.validate()

    def clear(self):
        for name, field in self.all_fields():

            # do nothing with required fields
            if field.is_required:
                continue

            if self._partial:
                self._set(name, NotSet)
            else:
                if field.flags['default'] == NotSet:
                    self._set(name, None)
                else:
                    self.to_default(name)

    def from_dict(self, data, cast=False):
        for name, field in self.all_fields():
            if name in data:
                self._set(name, data[name], cast)
        return self

    def to_dict(self):
        return dict(self.valid_items())

    def from_obj(self, obj, mapping=None, override=None, cast=False):
        """
        Maps the public attributes from a object to the resource fields based
        on identical names. Optional mapping parameter allows for discrepancies
        in naming with resource names being the key and the object attribute
        name to map to being the value. If explicit_mapping is True, only the
        attributes in the mapping dictionary will be copied. An optional
        override dictionary is used to set values explicitly.
        """
        if not override:
            override = {}

        for field_name, _ in self.all_fields():
            mapped_name = field_name

            # first attempt to find value in override
            value = override.get(field_name, NotSet)

            # if value isn't overridden, get it from the obj
            if value == NotSet:

                # Convert name to mapped name if available,
                # else use Resource's existing name
                if mapping and field_name in mapping:
                    mapped_name = mapping[field_name]

                    # None value indicates that we shouldn't map this field
                    if not mapped_name:
                        continue

                value = getattr(obj, mapped_name, NotSet)

                # If we still don't have a value, then skip this field
                if value is NotSet:
                    continue

            self._set(field_name, value, cast)

        return self

    def to_obj(self, obj, mapping=None, override=None):
        """
        Maps the fields from the resource to an object based on identical
        names. Optional mapping parameter allows for discrepancies in naming
        with resource names being the key and the object attribute name to map
        to being the value. If explicit_mapping is True, only the attributes in
        the mapping dictionary will be copied.
        """
        if not override:
            override = {}

        # Gets the names of all fields with set values, then adds any included
        # in the override. This allows even NotSet values to be overridden
        field_names = {n for n, f in self.valid_fields()} | set(override.keys())

        for field_name in field_names:
            mapped_name = field_name

            # first attempt to find value in override
            value = override.get(field_name, NotSet)

            # if value isn't overridden, get it from the field
            if value == NotSet:

                # Convert name to mapped name if available,
                # else use Resource's existing name
                if mapping and field_name in mapping:
                    mapped_name = mapping[field_name]

                    # None value indicates that we should skip this field
                    if not mapped_name:
                        continue

                # Ensure that target obj has an attribute with a matching name
                if not hasattr(obj, mapped_name):
                    continue

                # get the value from the field
                value = self._get(field_name)

                # if we still don't have a value, skip this field
                if value == NotSet:
                    continue

            # Set target obj value from Resource value
            setattr(obj, mapped_name, value)

        return obj

    @classmethod
    def all_fields(cls):
        return cls._fields.items()

    def valid_fields(self):
        return ((name, field) for name, field in self.all_fields() if self.is_set(name))

    def sorted_items(self):
        return ((n, self._get(n)) for n, v in sorted(self.all_fields(), key=lambda x: x[1].order_value))

    def items(self):
        return self._values.items()

    def valid_items(self):
        return ((name, field) for name, field in self.items() if self.is_set(name))

    def validate(self):
        if self._partial:
            self._validate_partial()
        else:
            self._validate_full()

    def _validate_full(self):
        for name, field in self._fields.items():
            if not (
                self.is_set(name)
                or field.is_readonly
                or field.is_immutable
            ):
                raise errors.ValidationError(
                    'Field {0} is NotSet, expected full resource.'
                    .format(field)
                )

    def _validate_partial(self):
        for name, field in self._fields.items():
            if not self.is_set(name) and field.is_required:
                raise errors.ValidationError(
                    'Field {0} is required, cannot be NotSet even on a '
                    'partial resource.'.format(field)
                )

    def to_default(self, field_name):
        self._set(field_name, self._fields[field_name].default)

    @property
    def is_partial(self):
        return self._partial

    def is_set(self, field_name):
        return self._get(field_name) != NotSet

    def _set(self, field_name, value, cast=False):
        field = self._fields[field_name]
        if value == NotSet and not (
            self._partial
            or field.is_readonly
            or field.is_immutable
        ):
            raise errors.ValidationError(
                'Attempted to set field {0} of a non-partial resource to '
                'NotSet'.format(field)
            )
        self._values[field_name] = field.validate(value, cast=cast)
        return self._values[field_name]

    def _get(self, field_name):
        return self._values[field_name]



"""
    thorium.logging
    ~~~~~~~~~~~~~

    Implements the logging support for Thorium.

"""

from logging import getLogger, StreamHandler, Formatter, getLoggerClass, DEBUG


def create_logger(logger_name, debug_on, debug_log_format):
    """Creates a logger for the given application.  This logger works
    similar to a regular Python logger but changes the effective logging
    level based on the application's debug flag.  Furthermore this
    function also removes all attached handlers in case there was a
    logger with the log name before.
    """
    Logger = getLoggerClass()

    class DebugLogger(Logger):
        def getEffectiveLevel(x):
            if x.level == 0 and debug_on:
                return DEBUG
            return Logger.getEffectiveLevel(x)

    class DebugHandler(StreamHandler):
        def emit(x, record):
            StreamHandler.emit(x, record) if debug_on else None

    handler = DebugHandler()
    handler.setLevel(DEBUG)
    handler.setFormatter(Formatter(debug_log_format))
    logger = getLogger(logger_name)
    # just in case that was not a new logger, get rid of all the handlers
    # already attached to it.
    del logger.handlers[:]
    logger.__class__ = DebugLogger
    logger.addHandler(handler)
    return logger



#make core objects available after import
from .thorium import Thorium

from .datastructures import NotSet

from .resources import Resource

from .endpoint import Endpoint

from .parameters import Parameters

from .routing import Route, RouteManager

from .request import Request

from .thoriumflask import ThoriumFlask

from . import fields




from . import errors


class Endpoint(object):
    """
    Responsible for the implementation of a single :class:`Resource`,
    The methods provide hooks which will be called into by the dispatcher
    during a request.

    :param request: A :class:`Request` object.
    :param response: A :class:`Response` object.
    """

    _authenticator_classes = None
    Resource = None

    def __init__(self, request, response):
        self.request = request
        self.response = response
        self._authenticators = []
        if self._authenticator_classes:
            self._authenticators = [a(self) for a in self._authenticator_classes]

    def authenticate(self, method):
        if self._authenticators:
            for auth in self._authenticators:
                auth.check_auth(method)

    def pre_request(self):
        pass

    def pre_request_detail(self):
        pass

    def pre_request_collection(self):
        pass

    def post_request(self):
        pass

    def get_detail(self):
        raise errors.MethodNotImplementedError()

    def get_collection(self):
        raise errors.MethodNotImplementedError()

    def post_detail(self):
        raise errors.MethodNotImplementedError()

    def post_collection(self):
        raise errors.MethodNotImplementedError()

    def put_detail(self):
        raise errors.MethodNotImplementedError()

    def put_collection(self):
        raise errors.MethodNotImplementedError()

    def delete_detail(self):
        raise errors.MethodNotImplementedError()

    def delete_collection(self):
        raise errors.MethodNotImplementedError()

    def patch_detail(self):
        raise errors.MethodNotImplementedError()

    def patch_collection(self):
        raise errors.MethodNotImplementedError()

    def options(self):
        raise errors.MethodNotImplementedError()





# -*- coding: utf-8 -*-

import traceback

from flask import Response as FlaskResponse, request as flaskrequest
from werkzeug.exceptions import BadRequest as WerkzeugBadRequest

from . import Thorium, errors
from .request import Request
from .resources import VALID_METHODS, VALID_QUERY_PARAMETERS
from .parameters import ParametersMetaClass
from .crossdomain_decorator import crossdomain


class ThoriumFlask(Thorium):

    def __init__(self, settings, route_manager, flask_app):
        self._flask_app = flask_app
        super(ThoriumFlask, self).__init__(
            settings=settings,
            route_manager=route_manager,
            debug=self._flask_app.debug,
        )

    def _bind_routes(self):
        routes = self._route_manager.get_all_routes()
        for r in routes:
            if r.path:
                fep = FlaskEndpoint(
                    dispatcher=r.dispatcher,
                    exception_handler=self.exception_handler,
                    flask_config=self._flask_app.config,
                )
                fep.__name__ = r.name
                self._flask_app.add_url_rule(
                    rule=r.path,
                    endpoint=r.name,
                    view_func=fep.endpoint_target,
                    methods=VALID_METHODS,
                )


class FlaskEndpoint(object):

    def __init__(self, dispatcher, exception_handler, flask_config):
        self.flask_config = flask_config
        self.dispatcher = dispatcher
        # should this just have a reference to the thorium object?
        self.exception_handler = exception_handler

    @crossdomain(origin='*')
    def endpoint_target(self, **kwargs):
        url = method = 'unknown'
        request = None
        try:
            request = self.build_request()
            url = request.url
            method = request.method
            response, serialized_body = self.dispatcher.dispatch(request)
            return FlaskResponse(
                response=serialized_body,
                headers=response.headers,
                status=response.status_code,
                content_type='application/json',
            )
        except errors.HttpErrorBase as e:
            error_body = self.exception_handler.handle_http_exception(
                http_error=e,
                request=request,
            )
            return FlaskResponse(
                response=error_body,
                status=e.status_code,
                headers=e.headers,
                content_type='application/json',
            )
        except Exception as e:
            error_body = self.exception_handler.handle_general_exception(
                url=url,
                method=method,
                e=e,
                request=request,
            )
            # if flask debug raise exception instead of returning json response
            if self.flask_config['DEBUG']:
                raise e
            return FlaskResponse(
                response=error_body,
                status=500,
                headers={},
                content_type='application/json',
            )

    def build_request(self):
        try:
            resource = None
            resources = []
            if flaskrequest.method.lower() in {'post', 'patch', 'put'}:
                if flaskrequest.mimetype == 'application/json':
                    json_data = flaskrequest.json or {}
                    partial = True if flaskrequest.method == 'PATCH' else False

                    # hack for single or list resources
                    if isinstance(json_data, list):
                        for i in json_data:
                            resources.append(self._create_resource(i, partial))
                    else:
                        resource = self._create_resource(json_data, partial)

                else:
                    raise errors.BadRequestError(
                        'Currently only json is supported, use application/'
                        'json mimetype')

            return Request(
                dispatcher=self.dispatcher,
                method=flaskrequest.method,
                identifiers=flaskrequest.view_args,
                query_params=self._build_parameters(),
                mimetype=flaskrequest.mimetype,
                resource=resource,
                resources=resources,
                url=flaskrequest.url,
            )
        except (errors.ValidationError, WerkzeugBadRequest) as e:
            raise errors.BadRequestError(message=e.args[0] if e.args else None)

    def _create_resource(self, data, partial):
        # override body data with url identifiers
        data.update(flaskrequest.view_args)
        if partial:
            resource = self.dispatcher.Resource.partial(data)
        else:
            resource = self.dispatcher.Resource(data)

        return resource

    def _build_parameters(self):
        if not self.dispatcher.Parameters:
            return None

        flask_params = flaskrequest.args.to_dict() if flaskrequest.args else {}

        if isinstance(self.dispatcher.Parameters, ParametersMetaClass):
            return self.dispatcher.Parameters.validate(flask_params)
        else:
            self._validate_no_extra_query_params(flask_params)
            flask_params.update(flaskrequest.view_args)
            params = self.dispatcher.Parameters.init_from_dict(
                data=flask_params,
                partial=True,
                cast=True,
            )
            params.sort = flask_params.get('sort')
            if not hasattr(params, 'offset'):
                params.offset = flask_params.get('offset')
            if not hasattr(params, 'limit'):
                params.limit = flask_params.get('limit')
            return params

    def _validate_no_extra_query_params(self, flask_params):
        param_fields = dict(self.dispatcher.Parameters.all_fields())
        for name, param in flask_params.items():
            if name not in param_fields and name not in VALID_QUERY_PARAMETERS:
                raise errors.ValidationError(
                    '{0} is not a supported query parameter.'.format(name)
                )



from datetime import timedelta
from flask import make_response, request, current_app
from functools import update_wrapper


def crossdomain(origin=None, methods=None, headers=None,
            max_age=21600, attach_to_all=True,
            automatic_options=True):
    if methods is not None:
        methods = ', '.join(sorted(x.upper() for x in methods))
    if headers is not None and not isinstance(headers, str):
        headers = ', '.join(x.upper() for x in headers)
    if not isinstance(origin, str):
        origin = ', '.join(origin)
    if isinstance(max_age, timedelta):
        max_age = max_age.total_seconds()

    def get_methods():
        if methods is not None:
            return methods

        options_resp = current_app.make_default_options_response()
        return options_resp.headers['allow']

    def decorator(f):
        def wrapped_function(*args, **kwargs):
            if automatic_options and request.method == 'OPTIONS':
                resp = current_app.make_default_options_response()
            else:
                resp = make_response(f(*args, **kwargs))
            if not attach_to_all and request.method != 'OPTIONS':
                return resp

            h = resp.headers

            h['Access-Control-Allow-Origin'] = request.headers['origin'] if 'origin' in request.headers else '*'
            h['Access-Control-Max-Age'] = str(max_age)
            h['Access-Control-Allow-Headers'] = 'accept, origin, content-type, X-EM-Token'
            h['Access-Control-Expose-Headers'] = 'Location, X-EM-Token'
            h['Access-Control-Allow-Credentials'] = 'true'
            if headers is not None:
                h['Access-Control-Allow-Headers'] = headers
            if h.get('Access-Control-Allow-Methods') is None:
                h['Access-Control-Allow-Methods'] = get_methods()
            return resp

        f.provide_automatic_options = False
        return update_wrapper(wrapped_function, f)
    return decorator



import json
import traceback
from flask import Response as FlaskResponse


class ValidationError(Exception):
    pass


class HttpErrorBase(Exception):
    status_code = None
    default_message = None

    def __init__(self, message=None, headers=None, code=None):
        self.headers = headers if headers else {}
        self.code = code
        super().__init__(message or self.default_message)


class BadRequestError(HttpErrorBase):
    status_code = 400
    default_message = 'Badly formed request.'


class UnauthorizedError(HttpErrorBase):
    status_code = 401
    default_message = 'Attempted action requires authentication.'


class ForbiddenError(HttpErrorBase):
    status_code = 403
    default_message = 'Supplied credentials not authorized for attempted action.'


class ResourceNotFoundError(HttpErrorBase):
    status_code = 404
    default_message = 'Expected resource was not found.'


class MethodNotAllowedError(HttpErrorBase):
    status_code = 405
    default_message = 'Attempted method not allowed on this resource.'


class ConflictError(HttpErrorBase):
    status_code = 409
    default_message = 'Request could not be processed because of a conflict.'


class InternalSeverError(HttpErrorBase):
    status_code = 500
    default_message = 'The server encountered an internal error or misconfiguration and was unable to complete your ' \
                      'request. Please try again later or contact the server administrator.'


class MethodNotImplementedError(HttpErrorBase):
    status_code = 500
    default_message = 'The attempted method was not found at this endpoint.'



# -*- coding: utf-8 -*-

from collections import OrderedDict
from operator import attrgetter

from . import errors


class Response(object):

    def __init__(self, request):
        self.meta = {
            'pagination': None,
        }
        self.headers = {}
        self.request = request
        self.error = None
        self.response_type = None
        self.status_code = self._set_status_code()

    def location_header(self, resource_id):
        ep = self.request.url
        if not ep.endswith('/'):
            ep += '/'
        ep += str(resource_id)
        self.headers['Location'] = ep

    def _set_status_code(self):
        if not self.request:  # Hacky
            return 500
        if self.request.method == 'POST':
            return 201
        elif self.request.method == 'DELETE':
            return 204
        else:
            return 200

    def get_response_data(self):
        raise NotImplementedError(
            'This method must be overridden by subclass.'
        )


class DetailResponse(Response):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.response_type = 'detail'
        self.resource = None

    def get_response_data(self):
        data = None
        if self.resource:
            data = OrderedDict(self.resource.sorted_items())
        return data


class CollectionResponse(Response):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.response_type = 'collection'
        self.resources = []
        self.sort = getattr(self.request.params, 'sort', None)
        self.meta['pagination'] = {
            'paginated': False,
            'limit': getattr(self.request.params, 'limit', None),
            'offset': getattr(self.request.params, 'offset', None),
            'record_count': 0,
        }

    def get_response_data(self):
        data = []
        if self.resources:
            self._sort()
            self._paginate()
            for res in self.resources:
                data.append(OrderedDict(res.sorted_items()))
        return data

    def _sort(self):
        if self.meta['pagination']['paginated']:
            return
        if self.sort:
            self.meta['sort'] = self.sort
            sort_by = self.sort.split(',')
            for field in reversed(sort_by):
                field, reverse = self._check_and_strip_first_char(field)
                self._validate_sort_field(field)
                not_sortable, sortable, = [], []
                for resource in self.resources:
                    if getattr(resource, field) is None:
                        not_sortable.append(resource)
                    else:
                        sortable.append(resource)
                sortable.sort(key=attrgetter(field), reverse=reverse)
                if reverse:
                    self.resources = sortable + not_sortable
                else:
                    self.resources = not_sortable + sortable

    def _paginate(self):
        if self.meta['pagination']['paginated']:
            return
        if  (self.meta['pagination']['offset'] is None or
             self.meta['pagination']['limit'] is None):
            return
        self._validate_offset_and_limit()
        if self.meta['pagination']['offset'] < 0:
            raise errors.BadRequestError('Offset cannot be negative.')
        if self.meta['pagination']['limit'] < 1:
            raise errors.BadRequestError('Limit must be greater than 1.')
        self.meta['offset'] = self.meta['pagination']['offset']
        self.meta['limit'] = self.meta['pagination']['limit']
        start = self.meta['offset']
        end = self.meta['pagination']['offset'] + self.meta['pagination']['limit']
        self.resources = self.resources[start:end]
        self.meta['pagination']['record_count'] = len(self.resources)
        self.meta['pagination']['next_page'] = start + len(self.resources)

    def _check_and_strip_first_char(self, field):
        reverse = field.startswith('-')
        field = field.lstrip('+-')
        return field, reverse

    def _validate_sort_field(self, field):
        if not hasattr(self.resources[0], field):
            raise errors.BadRequestError(
                'Cannot sort by field `{}`. It does not exist in the resource.'
                .format(field)
            )
        elif (isinstance(getattr(self.resources[0], field), dict) or
              isinstance(getattr(self.resources[0], field), list) or
              isinstance(getattr(self.resources[0], field), set)):
            raise errors.BadRequestError(
                'Cannot sort by field `{}`. Field type is not sortable.'
                .format(field)
            )

    def _validate_offset_and_limit(self):
        try:
            pagin = self.meta['pagination']
            if (isinstance(pagin['offset'], bool) or
                    isinstance(pagin['limit'], bool)):
                raise TypeError

            if pagin['offset'] is None and pagin['limit'] is None:
                return
            pagin['offset'] = int(pagin['offset'])
            pagin['limit'] = int(pagin['limit'])


        except (ValueError, TypeError):
            raise errors.BadRequestError(
                'Both `offset` and `limit` must be valid numbers. Could not '
                'cast offset of `{0}` or limit of `{1}`.'
                .format(self.meta['pagination']['offset'],
                        self.meta['pagination']['limit'])
            )


class ErrorResponse(Response):

    def __init__(self, http_error, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.response_type = 'error'
        self.status_code = http_error.status_code
        self.error = str(http_error)
        self.code = getattr(http_error, 'code', None)

    def get_response_data(self):
        return None
